/* https://mkdocstrings.github.io/crystal/styling.html#recommended-styles */

/* Indent and distinguish sub-items */
div.doc-contents:not(.first) {
  padding-left: 15px;
  border-left: 4px solid rgba(230, 230, 230);
}
